#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  utils.py
"""
Various tools
"""
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from collections.abc import Sequence, Set

# this package
from chemistry_tools.pubchem import API_BASE


def format_string(stringwithmarkup):
	"""
	Convert a PubChem formatted string into an HTML formatted string
	"""

	string = list(stringwithmarkup["String"])
	try:
		markup_list = stringwithmarkup["Markup"]
	except KeyError:
		markup_list = []

	for markup in markup_list:
		style = None
		start = markup["Start"]
		end = markup["Length"] + start - 1
		if markup["Type"] == "Italics":
			style = "i"
		# handle Other formats

		if style is None:
			print(markup)
			continue

		string[start] = f"<{style}>{string[start]}"
		string[end] = f"{string[end]}</{style}>"

	string = ''.join(string)

	return string


def _force_sequence_or_csv(value, name):
	if not value:
		raise ValueError(f"Please supply one or more {name}, either as a comma-separated string or a list of strings.")

	if isinstance(value, str):
		value = value.split(",")

	elif isinstance(value, int):
		value = [str(value)]

	elif isinstance(value, (Sequence, Set)):

		output = []

		for idx, val in enumerate(value):
			if isinstance(val, str):
				output.append(val)
			elif isinstance(val, int):
				output.append(str(val))
			else:
				raise ValueError(f"'{name}' must be either a comma-separated string or a Sequence of strings")

		value = output

	else:
		raise ValueError(f"'{name}' must be either a comma-separated string or a Sequence of strings")

	return value


def _make_base_url(namespace, identifier):

	identifier = _force_sequence_or_csv(identifier, "identifier")
	namespace = str(namespace)
	return f"{API_BASE}/compound/{namespace}/{','.join(identifier)}"



# -*- coding: utf-8 -*-
"""
test_Bond
~~~~~~~~~~~~~

Test Bond class

"""

import pytest

from chemistry_tools.pubchem.bond import Bond, BondType


@pytest.fixture(scope='module')
def b1():
	return Bond(1, 2, order=BondType.QUADRUPLE)


def test_bond(b1):
	assert b1.__repr__() == "Bond(1, 2, BondType.QUADRUPLE)"
	assert isinstance(b1.to_dict(), dict)
	assert b1.to_dict()["order"] == BondType.QUADRUPLE



#!/usr/bin/env python3
#
#  units.py
"""
The units module provides the following attributes:

- ``chempy.units.SI_base_registry``

together with some functions.

Currently `quantities <https://pypi.python.org/pypi/quantities>`_ is used as
the underlying package to handle units. If it is possible you should try to
only use the ``chempy.units`` module (since it is likely that ``ChemPy``
will change this backend at some point in the future). Therefore you should not
rely on any attributes of the ``Quantity`` instances (and rather use
getter & setter functions in `chempy.units`).

"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


# stdlib
from collections import OrderedDict
from functools import reduce
from operator import mul

# 3rd party
import numpy as np
import quantities

# this package
from chemistry_tools.dicts import ArithmeticDict
from chemistry_tools.utils import defaultnamedtuple


def is_quantity(arg):
	# this check works even if quantities is not installed.
	return bool(arg.__class__.__name__ == 'Quantity')


energy = ArithmeticDict(int, {'mass': 1, 'length': 2, 'time': -2})
volume = ArithmeticDict(int, {'length': 3})
concentration = {'amount': 1} - volume


def get_derived_unit(registry, key):
	"""
	Get the unit of a physcial quantity in a provided unit system.

	Parameters
	----------
	registry: dict (str: unit)
		mapping 'length', 'mass', 'time', 'current', 'temperature',
		'luminous_intensity', 'amount'. If registry is ``None`` the
		function returns 1.0 unconditionally.
	key: str
		one of the registry keys or one of: 'diffusivity', 'electricalmobility',
		'permittivity', 'charge', 'energy', 'concentration', 'density',
		'radiolytic_yield'.

	Examples
	--------
	>>> m, s = quantities.meter, quantities.second
	>>> get_derived_unit(SI_base_registry, 'diffusivity') == m**2/s
	True

	"""
	if registry is None:
		return 1.0

	derived = {
			'diffusivity': registry['length'] ** 2 / registry['time'],
			'electrical_mobility': registry['current'] * registry['time'] ** 2 / registry['mass'],
			'permittivity': (
					registry['current'] ** 2 * registry['time'] ** 4
					/ (registry['length'] ** 3 * registry['mass'])),
			'charge': registry['current'] * registry['time'],
			'energy': registry['mass'] * registry['length'] ** 2 / registry['time'] ** 2,
			'concentration': registry['amount'] / registry['length'] ** 3,
			'density': registry['mass'] / registry['length'] ** 3,
			}
	derived['diffusion'] = derived['diffusivity']  # 'diffusion' is deprecated
	derived['radiolytic_yield'] = registry['amount'] / derived['energy']
	derived['doserate'] = derived['energy'] / registry['mass'] / registry['time']
	derived['linear_energy_transfer'] = derived['energy'] / registry['length']

	try:
		return derived[key]
	except KeyError:
		return registry[key]


def unit_registry_to_human_readable(unit_registry):
	"""
	Serialization of a unit registry.
	"""

	if unit_registry is None:
		return None

	new_registry = {}
	integer_one = 1
	for k in SI_base_registry:
		if unit_registry[k] is integer_one:
			new_registry[k] = 1, 1
		else:
			dim_list = list(unit_registry[k].dimensionality)
			if len(dim_list) != 1:
				raise TypeError("Compound units not allowed: {}".format(
						dim_list))
			u_symbol = dim_list[0].u_symbol
			new_registry[k] = float(unit_registry[k]), u_symbol
	return new_registry


def _latex_from_dimensionality(dim):
	# see https://github.com/python-quantities/python-quantities/issues/148
	from quantities.markup import format_units_latex
	return format_units_latex(dim, mult=r'\\cdot')


def latex_of_unit(quant):
	"""
	Returns LaTeX reperesentation of the unit of a quantity

	Examples
	--------
	>>> print(latex_of_unit(1/quantities.kelvin))
	\\mathrm{\\frac{1}{K}}
	"""

	return _latex_from_dimensionality(quant.dimensionality).strip('$')


def unicode_of_unit(quant):
	"""
	Returns unicode reperesentation of the unit of a quantity

	Examples
	--------
	>>> print(unicode_of_unit(1/quantities.kelvin))
	1/K
	"""

	return quant.dimensionality.unicode


def html_of_unit(quant):
	"""
	Returns HTML reperesentation of the unit of a quantity

	Examples
	--------
	>>> print(html_of_unit(2*quantities.m**2))
	m<sup>2</sup>
	"""

	return quant.dimensionality.html


def unit_registry_from_human_readable(unit_registry):
	"""
	Deserialization of unit_registry.
	"""

	if unit_registry is None:
		return None

	new_registry = {}
	for k in SI_base_registry:
		factor, u_symbol = unit_registry[k]
		if u_symbol == 1:
			unit_quants = [1]
		else:
			unit_quants = list(quantities.Quantity(0, u_symbol).dimensionality.keys())

		if len(unit_quants) != 1:
			raise TypeError("Unknown UnitQuantity: {}".format(unit_registry[k]))
		else:
			new_registry[k] = factor * unit_quants[0]
	return new_registry


# Abstraction of underlying package providing units and dimensional analysis:

def is_unitless(expr):
	"""
		Returns ``True`` if ``expr`` is unitless, otherwise ``False``

	Examples
	--------
	>>> is_unitless(42)
	True
	>>> is_unitless(42*quantities.kilogram)
	False

	"""
	if hasattr(expr, 'dimensionality'):
		if expr == quantities.dimensionless:
			return True
		else:
			return expr.simplified.dimensionality == quantities.dimensionless.dimensionality
	if isinstance(expr, dict):
		return all(is_unitless(_) for _ in expr.values())
	elif isinstance(expr, (tuple, list)):
		return all(is_unitless(_) for _ in expr)
	return True


def unit_of(expr, simplified=False):
	"""
		Returns the unit of a quantity

	Examples
	--------
	>>> unit_of(42*pq.second) == unit_of(12*pq.second)
	True
	>>> unit_of(42)
	1

	"""
	if isinstance(expr, (tuple, list)):
		return unit_of(uniform(expr)[0], simplified)
	elif isinstance(expr, dict):
		return unit_of(list(uniform(expr).values())[0], simplified)

	try:
		if simplified:
			return expr.units.simplified
		else:
			return expr.units
	except AttributeError:
		return 1


def rescale(value, unit):
	try:
		return value.rescale(unit)
	except AttributeError:
		if unit == 1:
			return value
		else:
			raise


def to_unitless(value, new_unit=None):
	"""
	Nondimensionalization of a quantity.

	Parameters
	----------
	value: quantity
	new_unit: unit

	Examples
	--------
	>>> '%.1g' % to_unitless(1*quantities.metre, quantities.nm)
	'1e+09'
	>>> '%.1g %.1g' % tuple(to_unitless([1*quantities.m, 1*quantities.mm], quantities.nm))
	'1e+09 1e+06'

	"""
	integer_one = 1
	if new_unit is None:
		new_unit = quantities.dimensionless

	if isinstance(value, (list, tuple)):
		return np.array([to_unitless(elem, new_unit) for elem in value])
	elif isinstance(value, np.ndarray) and not hasattr(value, 'rescale'):
		if is_unitless(new_unit) and new_unit == 1 and value.dtype != object:
			return value
		return np.array([to_unitless(elem, new_unit) for elem in value])
	elif isinstance(value, dict):
		new_value = dict(value.items())  # value.copy()
		for k in value:
			new_value[k] = to_unitless(value[k], new_unit)
		return new_value
	elif isinstance(value, (int, float)) and new_unit is integer_one or new_unit is None:
		return value
	elif isinstance(value, str):
		raise ValueError("str not supported")
	else:
		try:
			try:
				result = (value * quantities.dimensionless / new_unit).rescale(quantities.dimensionless)
			except AttributeError:
				if new_unit == quantities.dimensionless:
					return value
				else:
					raise
			else:
				if result.ndim == 0:
					return float(result)
				else:
					return np.asarray(result)
		except TypeError:
			return np.array([to_unitless(elem, new_unit) for elem in value])


def uniform(container):
	"""
	Turns a list, tuple or dict with mixed units into one with uniform units.

	Parameters
	----------
	container : tuple, list or dict

	Examples
	--------
	>>> km, m = quantities.kilometre, quantities.metre
	>>> uniform(dict(a=3*km, b=200*m))  # doctest: +SKIP
	{'b': array(200.0) * m, 'a': array(3000.0) * m}

	"""
	if isinstance(container, (tuple, list)):
		unit = unit_of(container[0])
	elif isinstance(container, dict):
		unit = unit_of(list(container.values())[0])
		return container.__class__([(k, to_unitless(v, unit) * unit) for k, v in container.items()])
	else:
		return container
	return to_unitless(container, unit) * unit


def get_physical_dimensionality(value):
	if is_unitless(value):
		return {}

	_quantities_mapping = {
			quantities.UnitLength: 'length',
			quantities.UnitMass: 'mass',
			quantities.UnitTime: 'time',
			quantities.UnitCurrent: 'current',
			quantities.UnitTemperature: 'temperature',
			quantities.UnitLuminousIntensity: 'luminous_intensity',
			quantities.UnitSubstance: 'amount'
			}

	return {
			_quantities_mapping[k.__class__]: v for k, v
			in uniform(value).simplified.dimensionality.items()
			}


def _get_unit_from_registry(dimensionality, registry):
	return reduce(mul, [registry[k] ** v for k, v in dimensionality.items()])


def default_unit_in_registry(value, registry):
	_dimensionality = get_physical_dimensionality(value)

	if _dimensionality == {}:
		return 1

	return _get_unit_from_registry(_dimensionality, registry)


def unitless_in_registry(value, registry):
	_default_unit = default_unit_in_registry(value, registry)
	return to_unitless(value, _default_unit)


# NumPy like functions for compatibility:

def compare_equality(a, b):
	"""
	Returns True if two arguments are equal.

	Both arguments need to have the same dimensionality.

	Parameters
	----------
	a : quantity
	b : quantity

	Examples
	--------
	>>> km, m = quantities.kilometre, quantities.metre
	>>> compare_equality(3*km, 3)
	False
	>>> compare_equality(3*km, 3000*m)
	True

	"""
	# Work around for https://github.com/python-quantities/python-quantities/issues/146
	try:
		a + b
	except TypeError:
		# We might be dealing with e.g. None (None + None raises TypeError)
		try:
			len(a)
		except TypeError:
			# Assumed scalar
			return a == b
		else:
			if len(a) != len(b):
				return False
			return all(compare_equality(_a, _b) for _a, _b in zip(a, b))
	except ValueError:
		return False
	else:
		return a == b


def allclose(a, b, rtol=1e-8, atol=None):
	"""
		Analogous to ``numpy.allclose``. """
	try:
		d = abs(a - b)
	except Exception:
		try:
			if len(a) == len(b):
				return all(allclose(_a, _b, rtol, atol) for _a, _b in zip(a, b))
			else:
				return False
		except Exception:
			return False
	lim = abs(a) * rtol
	if atol is not None:
		lim += atol

	try:
		len(d)
	except TypeError:
		return d <= lim
	else:
		try:
			len(lim)
		except TypeError:
			return np.all([_d <= lim for _d in d])
		else:
			return np.all([_d <= _lim for _d, _lim in zip(d, lim)])


def logspace_from_lin(start, stop, num=50):
	"""
		Logarithmically spaced data points

	Examples
	--------
	>>> abs(logspace_from_lin(2, 8, num=3)[1] - 4) < 1e-15
	True

	"""
	unit = unit_of(start)
	start_ = np.log2(to_unitless(start, unit))
	stop_ = np.log2(to_unitless(stop, unit))
	return np.exp2(np.linspace(start_, stop_, num)) * unit


def _sum(iterable):
	try:
		result = next(iterable)
	except TypeError:
		result = iterable[0]
		for elem in iterable[1:]:
			result += elem
		return result
	else:
		try:
			while True:
				result += next(iterable)
		except StopIteration:
			return result
		else:
			raise ValueError("Not sure how this point was reached")


class Backend:
	"""
	Wrapper around modules such as numpy and math

	Instances of Backend wraps a module, e.g. `numpy` and ensures that
	arguments passed on are unitless, i.e. it raises an error if a
	transcendental function is used with quantities with units.

	Parameters
	----------
	underlying_backend : module, str or tuple of str
		e.g. 'numpy' or ('sympy', 'math')

	Examples
	--------
	>>> import math
	>>> km, m = quantities.kilometre, quantities.metre
	>>> math.exp(3*km) == math.exp(3*m)
	True
	>>> be = Backend('math')
	>>> be.exp(3*km)
	Traceback (most recent call last):
		...
	ValueError: Unable to convert between units of "km" and "dimensionless"
	>>> import numpy as np
	>>> np.sum([1000*pq.metre/pq.kilometre, 1])
	1001.0
	>>> be_np = Backend(np)
	>>> be_np.sum([[1000*pq.metre/pq.kilometre, 1], [3, 4]], axis=1)
	array([2., 7.])

	"""

	def __init__(self, underlying_backend=('numpy', 'math')):
		if isinstance(underlying_backend, tuple):
			for name in underlying_backend:
				try:
					self.be = __import__(name)
				except ImportError:
					continue
				else:
					break
			else:
				raise ValueError(f"Could not import any of {str(underlying_backend)}")
		elif isinstance(underlying_backend, str):
			self.be = __import__(underlying_backend)
		else:
			self.be = underlying_backend

	def __getattr__(self, attr):
		be_attr = getattr(self.be, attr)
		if callable(be_attr):
			return lambda *args, **kwargs: be_attr(*map(to_unitless, args), **kwargs)
		else:
			return be_attr


# TODO: decide whether to deprecate in favor of "number_to_scientific_latex"?
def format_string(value, precision='%.5g', tex=False):
	"""
	Formats a scalar with unit as two strings

	:param value: Value with unit
	:type value: float
	:param precision:
	:type precision: str, optional
	:param tex: Whether the string should be formatted for LaTex. Default :const:`False`
	:type tex: bool, optional

	:return:
	:rtype:

	Examples
	--------
	>>> print(' '.join(format_string(0.42*quantities.mol/quantities.decimetre**3)))
	0.42 mol/decimetre**3
	>>> print(' '.join(format_string(2/quantities.s, tex=True)))
	2 \\mathrm{\\frac{1}{s}}
	"""

	if tex:
		unit_str = latex_of_unit(value)
	else:
		from quantities.markup import config
		attr = 'unicode' if config.use_unicode else 'string'
		unit_str = getattr(value.dimensionality, attr)
	return precision % float(value.magnitude), unit_str


def concatenate(arrays, **kwargs):
	"""
	Patched version of numpy.concatenate

	Examples
	--------
	>>> from chemistry_tools.units import quantities
	>>> all(concatenate(([2, 3]*quantities.s, [4, 5]*quantities.s)) == [2, 3, 4, 5]*quantities.s)
	True

	"""
	unit = unit_of(arrays[0])
	result = np.concatenate([to_unitless(arr, unit) for arr in arrays], **kwargs)
	return result * unit


def fold_constants(arg):
	if hasattr(arg, 'dimensionality'):
		m = arg.magnitude
		d = 1
		for k, v in arg.dimensionality.items():
			if isinstance(k, quantities.UnitConstant):
				m = m * k.simplified ** v
			else:
				d = d * k ** v
		return m * d
	else:
		return arg


def _sanity_check_quantities(pq):
	# See https://github.com/python-quantities/python-quantities/pull/116
	a = pq.UncertainQuantity([1, 2], pq.m, [.1, .2])
	assert (-a).uncertainty[0] == (a * -1).uncertainty[0]

	# See https://github.com/python-quantities/python-quantities/pull/126
	assert (3 * pq.m) ** 0 == 1 * pq.dimensionless


_sanity_check_quantities(quantities)

# Additional units to complement quantities
per100eV = quantities.UnitQuantity(
		'per_100_eV',
		1 / (100 * quantities.eV * quantities.constants.Avogadro_constant),
		u_symbol='(100eV)**-1')
dm = decimetre = quantities.UnitQuantity('decimetre', quantities.m / 10.0, u_symbol='dm')
m3 = quantities.metre ** 3
dm3 = decimetre ** 3
cm3 = quantities.centimetre ** 3
nanomolar = quantities.UnitQuantity('nM', 1e-6 * quantities.mole / m3, u_symbol='nM')
molal = quantities.UnitQuantity('molal', quantities.mole / quantities.kg, u_symbol='molal')
micromole = quantities.UnitQuantity('micromole', quantities.mole / 1e6, u_symbol='μmol')
nanomole = quantities.UnitQuantity('nanomole', quantities.mole / 1e9, u_symbol='nmol')
kilojoule = quantities.UnitQuantity('kilojoule', 1e3 * quantities.joule, u_symbol='kJ')
kilogray = quantities.UnitQuantity('kilogray', 1e3 * quantities.gray, u_symbol='kGy')
perMolar_perSecond = 1 / quantities.molar / quantities.s
umol_per_J = quantities.umol / quantities.joule

# unit registry data and logic:
SI_base_registry = {
		'length': quantities.metre,
		'mass': quantities.kilogram,
		'time': quantities.second,
		'current': quantities.ampere,
		'temperature': quantities.kelvin,
		'luminous_intensity': quantities.candela,
		'amount': quantities.mole
		}


dimension_codes = OrderedDict(zip(
		'length mass time current temperature amount'.split(),  # not considering luminous_intensity
		'L M T I Θ N'.split()
		))


class DimensionalitySI(defaultnamedtuple('DimensionalitySIBase', dimension_codes.keys(), (0,) * len(dimension_codes))):

	def __mul__(self, other):
		return self.__class__(*(x + y for x, y in zip(self, other)))

	def __truediv__(self, other):
		return self.__class__(*(x - y for x, y in zip(self, other)))

	def __pow__(self, exp):
		return self.__class__(*(x * exp for x in self))


base_registry = {name: DimensionalitySI(**{name: 1}) for name in dimension_codes}



#!/usr/bin/env python
#
#  atom.py
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from itertools import zip_longest

# this package
from chemistry_tools.elements import ELEMENTS
from chemistry_tools.pubchem.errors import ResponseParseError


class Atom:
	"""Class to represent an atom in a :class:`~pubchempy.Compound`."""

	def __init__(self, aid, number, x=None, y=None, z=None, charge=0):
		"""Initialize with an atom ID, atomic number, coordinates and optional change.

		:param int aid: Atom ID
		:param int number: Atomic number
		:param float x: X coordinate.
		:param float y: Y coordinate.
		:param float z: (optional) Z coordinate.
		:param int charge: (optional) Formal charge on atom.
		"""

		self.aid = aid
		# The atom ID within the owning Compound.
		self.number = number
		# The atomic number for this atom.
		self.x = x
		# The x coordinate for this atom.
		self.y = y
		# The y coordinate for this atom.
		self.z = z
		# The z coordinate for this atom. Will be ``None`` in 2D Compound records.
		self.charge = charge
		# The formal charge on this atom.

	def __repr__(self):
		return f'Atom({self.aid}, {self.element})'

	def __eq__(self, other):
		return (
				isinstance(other, type(self))
				and self.aid == other.aid
				and self.element == other.element
				and self.x == other.x
				and self.y == other.y
				and self.z == other.z
				and self.charge == other.charge
				)

	@property
	def element(self):
		"""
		The element symbol for this atom.
		"""

		return ELEMENTS[self.number].symbol

	def to_dict(self):
		"""
		Return a dictionary containing Atom data.
		"""

		data = {'aid': self.aid, 'number': self.number, 'element': self.element}

		for coord in {'x', 'y', 'z'}:
			if getattr(self, coord) is not None:
				data[coord] = getattr(self, coord)

		if self.charge != 0:
			data['charge'] = self.charge

		return data

	def set_coordinates(self, x, y, z=None):
		"""
		Set all coordinate dimensions at once.
		"""

		self.x = x
		self.y = y
		self.z = z

	@property
	def coordinate_type(self):
		"""
		Returns whether this atom has 2D or 3D coordinates.

		:rtype: str
		"""

		if self.z is None:
			return '2d'
		else:
			return '3d'


def parse_atoms(atoms_dict, coords_dict=None):
	"""

	:param atoms_dict:
	:type atoms_dict: dict
	:param coords_dict:
	:type coords_dict: dict or None

	:return:
	:rtype: dict
	"""

	atoms = {}

	# Create atoms
	aids = atoms_dict['aid']
	elements = atoms_dict['element']

	if not len(aids) == len(elements):
		raise ResponseParseError('Error parsing atom elements')

	for aid, element in zip(aids, elements):
		atoms[aid] = Atom(aid=aid, number=element)

	# Add coordinates
	if coords_dict:
		coord_ids = coords_dict[0]['aid']

		xs = coords_dict[0]['conformers'][0]['x']
		ys = coords_dict[0]['conformers'][0]['y']
		zs = coords_dict[0]['conformers'][0].get('z', [])

		if not len(coord_ids) == len(xs) == len(ys) == len(atoms) or (zs and not len(zs) == len(coord_ids)):
			raise ResponseParseError('Error parsing atom coordinates')

		for aid, x, y, z in zip_longest(coord_ids, xs, ys, zs):
			atoms[aid].set_coordinates(x, y, z)

	# Add charges
	if 'charge' in atoms_dict:
		for charge in atoms_dict['charge']:
			atoms[charge['aid']].charge = charge['value']

	return atoms



# -*- coding: utf-8 -*-
"""
test_compound3d
~~~~~~~~~~~~~~~

Test compound object with 3D record.

"""

# 3rd party
import pytest

# this package
from chemistry_tools.pubchem.compound import Compound


@pytest.fixture
def c3d():
	"""Compound CID 1234, 3D."""
	return Compound.from_cid(1234, record_type='3d')


def test_coordinate_type(c3d):
	assert c3d.coordinate_type == '3d'


def test_atoms(c3d):
	assert len(c3d.atoms) == 75
	assert set(a.element for a in c3d.atoms) == {'C', 'H', 'O', 'N'}
	assert set(c3d.elements) == {'C', 'H', 'O', 'N'}


def test_coordinates(c3d):
	for a in c3d.atoms:
		assert isinstance(a.x, (float, int))
		assert isinstance(a.y, (float, int))
		print(a.z)
		assert isinstance(a.z, (float, int))




#!/usr/bin/env python3
#
#  test_formula.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#


# stdlib
import decimal

# 3rd party
import pytest
from mathematical.utils import rounders

from chemistry_tools.formulae import Formula, Species
from chemistry_tools.elements import isotope_data, D, O


def test_formula():
	# From string
	f1 = Formula.from_string("(C6H5)2NH")
	f2 = Formula.from_string("C12H11N")
	assert f1 == f2

	# from dict
	f3 = Formula({"C": 12, "H": 11, "N": 1})
	assert f1 == f3
	assert f2 == f3

	# from kwargs
	f4 = Formula.from_kwargs(C=12, H=11, N=1)
	assert f1 == f4
	assert f2 == f4
	assert f3 == f4

	# from composition
	f5 = Formula(f1)
	assert f5 == f1
	assert f5 == f2

	s = Formula.from_string('H+')
	assert s.charge == 1
	assert s.mass == 1.007941


def test_charged_formula():
	# From string
	f1 = Formula.from_string("(C6H5)2NH+")
	f2 = Formula.from_string("C12H11N", charge=1)
	assert f1 == f2

	# from dict
	f3 = Formula({"C": 12, "H": 11, "N": 1}, charge=1)
	assert f3 == {"C": 12, "H": 11, "N": 1}
	assert f1 == f3
	assert f2 == f3

	# from kwargs
	f4 = Formula.from_kwargs(C=12, H=11, N=1, charge=1)
	assert f1 == f4
	assert f2 == f4
	assert f3 == f4

	# from composition
	f5 = Formula(f1)
	assert f5 == f1
	assert f5 == f2

	for charge in [1, 2, 3]:
		with pytest.raises(ValueError):
			Formula.from_string(f'BCHFKOH+{charge:d}', charge + 1)


def test_isotope_formula():
	# from dict
	f3 = Formula({"C[12]": 10, "C[13]": 2, "H": 11, "N": 1})
	assert dict(f3) == {"[12C]": 10, "[13C]": 2, "H": 11, "N": 1}

	# from composition
	f5 = Formula(f3)
	assert f5 == f3


def test_formula_sum():
	# Test sum of Formula objects.
	assert Formula.from_string('C6H12O6') + Formula.from_string('H2O') == {"C": 6, "H": 14, "O": 7}


def test_Composition_sub():
	# Test subtraction of Composition objects
	assert {} - Formula.from_string('C6H12O6') == {"C": -6, "H": -12, "O": -6}


def test_Composition_mul():
	# Test multiplication of Composition by integers
	f1 = Formula.from_string("H2O2")
	assert f1 == {"H": 2, "O": 2}
	assert 2 * f1 == {"H": 4, "O": 4}
	assert f1 * 2 == {"H": 4, "O": 4}


def test_calculate_mass():
	# Calculate mass by a formula.
	assert rounders(Formula.from_string('(C6H5)2NH').monoisotopic_mass, "0.000000") == decimal.Decimal("169.089149")

	# Calculate average mass / molecular weight by a formula.
	assert rounders(Formula.from_string('(C6H5)2NH').average_mass, "0.00") == decimal.Decimal("169.22")
	assert rounders(Formula.from_string('(C6H5)2NH').average_mass, "0.00") == decimal.Decimal("169.22")
	assert Formula.from_string('(C6H5)2NH').average_mass == Formula.from_string('(C6H5)2NH').average_mass

	# mz
	assert Formula.from_string('C12H13N+').get_mz() == Formula.from_string('C12H13N', charge=1).average_mass


@pytest.mark.parametrize("charge", [1, 2, 3, 4])
def test_calculate_mz(charge):
	# Calculate m/z of an ion.
	assert Formula.from_string('C12H13N+').exact_mass == Formula.from_string('C12H13N', charge=1).exact_mass
	# perhaps math.isclose()
	# assert Formula.from_string('C12H13N+').mass == Formula.from_string('C12H13N', charge=1).exact_mass
	# assert Formula.from_string('C12H13N+').exact_mass == Formula.from_string('C12H13N', charge=1).mass
	# assert Formula.from_string('C12H13N+').mass == Formula.from_string('C12H13N', charge=1).mass

	assert Formula.from_string('C12H13N+').get_mz() == Formula.from_string('C12H13N', charge=1).get_mz()
	assert Formula.from_string('C12H13N+').mz == Formula.from_string('C12H13N', charge=1).get_mz()
	assert Formula.from_string('C12H13N+').get_mz() == Formula.from_string('C12H13N', charge=1).mz


def test_most_probable_isotopic_composition():
	assert (
			Formula.from_string('F').most_probable_isotopic_composition()
			== (Formula({'F[19]': 1, 'F[18]': 0}), 1.0)
			)

	Br2 = Formula.from_string('Br2')
	assert Br2.most_probable_isotopic_composition()[0] == Formula({'Br[79]': 1, 'Br[81]': 1})
	assert rounders(Br2.most_probable_isotopic_composition()[1], "0.000") == decimal.Decimal("0.5")

	C6Br6 = Formula.from_string('C6Br6')
	assert C6Br6.most_probable_isotopic_composition()[0] == Formula({'C[12]': 6, 'C[13]': 0, 'Br[79]': 3, 'Br[81]': 3})
	assert rounders(C6Br6.most_probable_isotopic_composition()[1], "0.000") == decimal.Decimal("0.293")

	assert (
			Formula.from_string('F10').most_probable_isotopic_composition()
			== (Formula({'F[19]': 10, }), 1.0)
			)

	assert Formula.from_string('CF4').most_probable_isotopic_composition(
			elements_with_isotopes=['F']) == (Formula({'C': 1, 'F[19]': 4}), 1.0)


def test_isotope_data():
	print(list(type((g[0][1]) - 1) for g in isotope_data.values()))
	assert all(abs(g[0][1] - 1) < 1e-6 for g in isotope_data.values())
	for g in isotope_data.values():
		s = sum(p[1] for num, p in g.items() if num)
		assert abs(s - 1) < 1e-6 or abs(s) < 1e-6


def test_iter_isotopologues():

	iter_isotopologues = Formula.from_string("C6Br6").iter_isotopologues()
	assert len(list(iter_isotopologues)) == 49

	iter_isotopologues = Formula.from_string("C6Br6").iter_isotopologues(
			elements_with_isotopes='Br')
	assert len(list(iter_isotopologues)) == 7


# TODO:
def test_iter_isotopologues_with_abundances():

	for state, abundance in Formula.from_string("BCHFKO").iter_isotopologues(
			elements_with_isotopes='F', report_abundance=True):
		assert state
		assert abundance


@pytest.fixture(scope="module")
def Br2():
	return Formula.from_string('Br2')


@pytest.fixture(scope="module")
def C6Br6():
	return Formula.from_string('C6Br6')


def test___str__(Br2, C6Br6):
	assert str(Br2) == "Formula({'Br': 2})"
	assert str(C6Br6) == "Formula({'C': 6, 'Br': 6})"


def test___repr__(Br2, C6Br6):
	assert repr(Br2) == "Formula({'Br': 2})"
	assert repr(C6Br6) == "Formula({'C': 6, 'Br': 6})"


@pytest.mark.parametrize("other", ["Br2", "abc", 123, 12, 34, [1, 2, 3]])
def test_unsupported_equals(Br2, other):
	assert Br2 != other


def test_Species():
	s = Species.from_string('H2O')
	assert s.phase is None
	assert Species.from_string('CO2(g)').phase == "g"
	assert Species.from_string('CO2(aq)').phase == "aq"
	assert Species.from_string('NaCl(s)').phase == "s"
	assert Species.from_string('H2O(l)').phase == "l"


def test_properties():
	f = Formula.from_string('C6H12O6')  # Sugar
	assert f.hill_formula == "C6H12O6"
	assert f.empirical_formula == "CH2O"
	assert f.n_atoms == 24
	assert f.n_elements == 3

	f = Formula.from_string('D2O')  # heavy water
	assert f.hill_formula == "D2O"
	assert f.empirical_formula == "D2O"
	assert f.monoisotopic_mass == 20.02311817516
	assert f.mass == 20.0276085556
	assert f.n_atoms == 3
	assert f.n_elements == 2
	assert D.nominalmass == 2
	assert O.nominalmass == 16


@pytest.mark.parametrize("formula, mass, exact_mass", [
		("C[12]", "12.00000", "12.00000"),
		("[12C]", "12.00000", "12.00000"),
		("[C12]", "12.00000", "12.00000"),
		('C[12]C', "24.01074", "24.00000"),
		('C[C12]', "24.01074", "24.00000"),
		('Co(Bpy)(CO)4', "327.15811", "326.98160"),
		('CH3CH2Cl', "64.51408", "64.00798"),
		('C2H5Cl', "64.51408", "64.00798"),
		('C1000H1000', "13018.68100", "13007.82503"),
		('Ru2(CO)8', "426.22116", "427.76802"),
		('RuClH(CO)(PPh3)3', "952.39958", "952.14935"),
		('PhSiMe3', "150.29333", "150.08648"),
		('C9H14Si', "150.29333", "150.08648"),
		('Ph(CO)C(CH3)3', "162.22872", "162.10447"),
		('C11H14O', "162.22872", "162.10447"),
		('HGlyGluTyrOH', "367.35455", "367.13795"),
		('C16H21N3O7', "367.35455", "367.13795"),
		('HCysTyrIleGlnAsnCysProLeuNH2', "952.15329", "951.43064"),
		('C41H65N11O11S2', "952.15329", "951.43064"),
		('HCysp(Trt)Tyrp(Tbu)IleGlnp(Trt)Asnp(Trt)ProLeuGlyNH2', "1689.11406", "1687.83417"),
		('C101H113N11O11S', "1689.11406", "1687.83417"),
		('C116H148N46O73P12', "3726.37115", "3724.61342"),
		('C47H83N15O16S', "1146.31971", "1145.58629"),
		('CDCl3', "120.38354", "118.92066"),
		('C[2H]Cl3', "120.38354", "118.92066"),
		('[13C]Cl4', "154.81495", "152.87877"),
		('[C13]Cl4', "154.81495", "152.87877"),
		('C5(PhBu(EtCHBr)2)3', "1194.60962", "1188.12038"),
		('C53H78Br6', "1194.60962", "1188.12038"),
		# TODO: ('AgCuRu4(H)2[CO]12{PPh3}2', "1438.4022", "1439.588960"),
		('C48H32AgCuO12P2Ru4', "1438.40422", "1439.58899"),
		('C6H8ClN', "129.58757", "129.03453"),
		('PhNH2.HCl', "129.58757", "129.03453"),
		('NH3.BF3', "84.83674", "85.03106"),
		('BF3H3N', "84.83674", "85.03106"),
		('CuH10O9S', "249.68485", "248.93415"),
		('CuSO4.5H2O', "249.68485", "248.93415"),
		])
def test_masses(formula, mass, exact_mass):
	f = Formula.from_string(formula)
	print(f)
	assert rounders(f.mass, "0.00000") == decimal.Decimal(mass)
	assert rounders(f.exact_mass, "0.00000") == decimal.Decimal(exact_mass)


@pytest.mark.parametrize("formula_1, formula_2", [
		('CH3CH2Cl', 'C2H5Cl'),
		('RuClH(CO)(PPh3)3', 'C55H46ClOP3Ru'),
		('PhSiMe3', 'C9H14Si'),
		('Ph(CO)C(CH3)3', 'C11H14O'),
		('HGlyGluTyrOH', 'C16H21N3O7'),
		('HCysTyrIleGlnAsnCysProLeuNH2', 'C41H65N11O11S2'),
		('HCysp(Trt)Tyrp(Tbu)IleGlnp(Trt)Asnp(Trt)ProLeuGlyNH2', 'C101H113N11O11S'),
		# TODO: ('CDCl3', 'C[2H]Cl3'), needs special case
		('[13C]Cl4', '[13C]Cl4'),
		('C5(PhBu(EtCHBr)2)3', 'C53H78Br6'),
		('PhNH2.HCl', 'C6H8ClN'),
		('NH3.BF3', 'BF3H3N'),
	])
def test_equivalent(formula_1, formula_2):
	f1 = Formula.from_string(formula_1)
	f2 = Formula.from_string(formula_2)
	assert f1 == f2


@pytest.mark.parametrize("formula_1, formula_2", [
		('CuSO4.5H2O', 'CuH10O9S'),
		('C1000H1000', 'CH'),
		('Ru2(CO)8', 'C4O4Ru'),
	])
def test_empirical_formula(formula_1, formula_2):
	f1 = Formula.from_string(formula_1)
	f2 = Formula.from_string(formula_2)
	assert f1.empirical_formula == f2.empirical_formula
	assert f1.empirical_formula == formula_2


@pytest.mark.parametrize("formula", [
		'', '()', '2', 'a', '(a)',
		# TODO: 'C:H', 'H:', 'C[H', 'H)2',  # failing
		'A', 'Aa', '2lC', '1C', '()0',
		'H0', '(H)0C', 'Ox: 0.26, 30Si: 0.74',
		"Hey", "O2Hey", "HeyO2",
	])
def test_invalid_formulae(formula):
	with pytest.raises(ValueError):
		print(Formula.from_string(formula))


@pytest.mark.parametrize("formula, data", [
		("H2O", {"H": 2, "O": 1}),
		("[2H]2O", {"[2H]": 2, "O": 1}),
		("CH3COOH", {"C": 2, "O": 2, "H": 4}),
		("EtOH", {"C": 2, "O": 1, "H": 6}),
		("CuSO4.5H2O", {"Cu": 1, "O": 9, "H": 10, "S": 1}),
		("(COOH)2", {"C": 2, "O": 4, "H": 2}),
		# TODO: ("AgCuRu4(H)2[CO]12{PPh3}2", {}),
		# TODO: ("CGCGAATTCGCG", {}),
		# TODO: ("MDRGEQGLLK", {}),
		])
def test_parsing(formula, data):
	assert Formula.from_string(formula) == data



#!/usr/bin/env python3
#
#  names.py
"""
Functions for working with IUPAC names for chemicals
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import re

# this package
from chemistry_tools.constants import prefixes

multiplier_regex = "*".join([f"({prefix})" for prefix in prefixes.values()])

re_strings = [
		r"((\d+),?)+(\d+)-",
		fr"{multiplier_regex}*",
		r"nitro",
		r"phenyl",
		r"aniline",
		r"anisole",
		r"benzene",
		r"centralite",
		r"formamide",
		r"glycerine",
		r"nitrate",
		r"glycol",
		r"phthalate",
		r"picrate",
		r"toluene",
		r" ",
		]


def get_IUPAC_parts(string):
	split_points = set()

	for regex in re_strings:
		for match in list(re.finditer(regex, string.lower())):
			start, end = match.span()
			if start != end:
				split_points.add(start)
				split_points.add(end)

	split_points.discard(0)
	split_points = sorted(split_points)
	start_point = 0

	string_chars = list(string)
	elements = []
	for point in split_points:
		elements.append("".join(string_chars[start_point:point]))
		start_point = point

	elements.append("".join(string_chars[start_point:]))

	while not elements[-1]:
		elements = elements[:-1]

	return elements


#
# from string import ascii_letters
# alphabet = ascii_letters + "0123456789" + ",'" + "- " + '!"#$%&()*+./:;<=>?@[\\]^_`{|}~'


def sort_IUPAC_names(iupac_names):
	"""
	Sort the list of IUPAC names into order.

	:param iupac_names:
	:type iupac_names: List[str]

	:return:
	:rtype: List[str]
	"""

	sort_order = get_IUPAC_sort_order(iupac_names)

	# return [iupac_names[split_names.index(name)] for name in sorted_names]
	return sorted(iupac_names, key=lambda x: sort_order[x])


def get_IUPAC_sort_order(iupac_names):
	"""
	Returns the order the names should be sorted in.

	Useful when sorting arrays containing data in addition to the name.
	e.g.
	>>> sort_order = get_IUPAC_sort_order([row[0] for row in data])
	>>> sorted_data = sorted(data, key=lambda row: sort_order[row[0]])

	where row[0] would be the name of the compound

	:param iupac_names:
	:type iupac_names: List[str]

	:return:
	:rtype: List[int]
	"""

	split_names, sorted_names = _get_split_and_sorted_lists(iupac_names)

	sort_order = {}
	for index, name in enumerate(sorted_names):
		sort_order[iupac_names[split_names.index(name)]] = index

	return sort_order


def get_sorted_parts(iupac_names):
	"""
	Returns parts of the IUPAC names sorted into order.
	# The parts are in reverse order (i.e. diphenylamine becomes ["amine", "phenyl", "di"]).

	:param iupac_names:
	:type iupac_names: List[str]

	:return:
	:rtype: List[List[str]]
	"""

	split_names, sorted_names = _get_split_and_sorted_lists(iupac_names)

	return [split_names[split_names.index(name)] for name in sorted_names]


def _get_split_and_sorted_lists(iupac_names):
	split_names = []

	for name in iupac_names:
		split_name = get_IUPAC_parts(name.lower())

		if split_name[0].lower() in prefixes.values():
			# no positional information at beginning
			split_name = [" ", *split_name]

		split_names.append(split_name[::-1])

	sorted_names = sorted(split_names)

	return split_names, sorted_names


def sort_array_by_name(array, name_col=0):
	"""
	Sort a list of lists by the IUPAC name in each row.

	:param array:
	:type array: List[List]
	:param name_col: The index of the column containing the IUPAC names
	:type name_col: int

	:return:
	:rtype:
	"""

	names = [row[name_col] for row in array]
	sort_order = get_IUPAC_sort_order(names)
	sorted_array = sorted(array, key=lambda row: sort_order[row[0]])
	return sorted_array







# -*- coding: utf-8 -*-
"""
test_Errors
~~~~~~~~~~~~~

Test Errors.

"""


import pytest

from chemistry_tools.pubchem.compound import Compound
from chemistry_tools.pubchem.lookup import get_compounds
from chemistry_tools.pubchem.errors import BadRequestError, NotFoundError


def test_invalid_identifier():
	"""BadRequestError should be raised if identifier is not a positive integer."""
	with pytest.raises(BadRequestError):
		Compound.from_cid('aergaerhg')
	with pytest.raises(NotFoundError):
		get_compounds('srthrthsr')


def test_notfound_identifier():
	"""
	NotFoundError should be raised if identifier is a positive integer but record doesn't exist.
	"""

	with pytest.raises(NotFoundError):
		Compound.from_cid(999999999)
	with pytest.raises(NotFoundError):
		get_compounds(999999999, "cid")



#!/usr/bin/env python3
#
#  full_record.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# this package
from chemistry_tools.pubchem.enums import PubChemNamespace
from chemistry_tools.pubchem.pug_rest import _do_rest_get
from chemistry_tools.pubchem.properties import _parse_record_property


def parse_full_record(record):
	parsed_records = []

	for compound in record["PC_Compounds"]:
		cid = compound["id"]["id"]["cid"]
		# print(cid)

		counts = compound["count"]
		# pprint(counts)

		properties = []
		for prop in compound["props"]:
			prop = _parse_record_property(prop)
			properties.append(prop)

		if "bonds" in compound:
			bonds = compound["bonds"]
		else:
			bonds = {}

		if "charge" in compound:
			charge = compound["charge"]
		else:
			charge = 0

		parsed_records.append(dict(
				atoms=compound["atoms"],
				bonds=bonds,
				charge=charge,
				coords=compound["coords"],
				properties=properties,
				cid=cid,
				counts=counts
				))

	return parsed_records


def rest_get_full_record(identifier, namespace=PubChemNamespace.name, record_type="2d", **kwargs):
	"""
	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:type identifier: str, Sequence[str]
	:param namespace: The type of identifier to look up. Valid values are in :class:`PubChemNamespace`
	:type namespace: PubChemNamespace, optional
	:type namespace: str, optional
	:param **kwargs: Optional arguments that ``json.loads`` takes.
	:raises ValueError: If the response body does not contain valid json.

	:return: Parsed json data
	:rtype: dict
	"""

	return _do_rest_get(namespace, identifier, record_type=record_type).json(**kwargs)




# -*- coding: utf-8 -*-
"""
test_properties
~~~~~~~~~~~~~~~

Test properties requests.

"""


from chemistry_tools.pubchem.properties import get_properties
from chemistry_tools.pubchem.synonyms import get_synonyms
from chemistry_tools.pubchem.utils import format_string


def test_properties():
	results = get_properties('tris-(1,10-phenanthroline)ruthenium', ['IsomericSMILES', 'InChIKey'], 'name')
	assert len(results) > 0
	for result in results:
		assert 'CID' in result
		assert 'IsomericSMILES' in result
		assert 'InChIKey' in result


def test_underscore_properties():
	"""
	Properties can also be specified as underscore-separated words, rather than CamelCase.
	"""

	results = get_properties('tris-(1,10-phenanthroline)ruthenium', ['IsomericSMILES', 'MolecularWeight'], 'name')
	assert len(results) > 0
	for result in results:
		assert 'CID' in result
		assert 'IsomericSMILES' in result
		assert 'MolecularWeight' in result


def test_comma_string_properties():
	"""
	Properties can also be specified as a comma-separated string, rather than a list.
	"""

	results = get_properties('tris-(1,10-phenanthroline)ruthenium', 'IsomericSMILES,InChIKey,MolecularWeight', 'name')
	assert len(results) > 0
	for result in results:
		assert 'CID' in result
		assert 'IsomericSMILES' in result
		assert 'MolecularWeight' in result
		assert 'InChIKey' in result


def test_synonyms():
	results = get_synonyms('C1=CC2=C(C3=C(C=CC=N3)C=C2)N=C1', 'smiles')
	assert len(results) > 0
	for result in results:
		assert 'CID' in result
		assert 'synonyms' in result
		assert isinstance(result['synonyms'], list)
		assert len(result['synonyms']) > 0


stringwithmarkup = {
		'String': 'N-phenylaniline',
		'Markup': [{'Start': 0, 'Length': 1, 'Type': 'Italics'}]
		}


def test_format_string():
	html_string = format_string(stringwithmarkup)
	assert isinstance(html_string, str)
	assert html_string == '<i>N</i>-phenylaniline'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  parser.py
"""
Functions and constants for parsing formulae
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


# stdlib
import re
from collections import defaultdict
from functools import lru_cache
from string import ascii_lowercase, ascii_uppercase

# 3rd party
import pyparsing

# this package
from chemistry_tools.elements import ELEMENTS
from ._parser_core import _formula_to_parts, _get_charge, _get_leading_integer, _parse_multiplicity
from .latex import _latex_mapping


_atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
_formula = fr'^({_atom})*$'
relative_atomic_masses = [element.mass for element in ELEMENTS]


# Construct regular expression to match all elements, plus D and T
element_re_dict = {}

for element in (ELEMENTS.symbols + ["D", "T"]):
	if len(element) == 1:
		if element in element_re_dict:
			element_re_dict[element].append("?")
		else:
			element_re_dict[element] = ["?"]
	else:
		upper, lower = list(element)
		if upper in element_re_dict:
			element_re_dict[upper].append(lower)
		else:
			element_re_dict[upper] = [lower]

invalid_uppers = list(ascii_uppercase)
invalid_lowers = {upper: list(ascii_lowercase) for upper in list(ascii_uppercase)}
invalid_re = []

element_re = []
for upper, lowers in element_re_dict.items():
	lowers = sorted(lowers)

	if lowers == ["?"]:
		element_re.append(upper)
		invalid_uppers.remove(upper)
	elif len(lowers) == 1:
		element_re.append(f"{upper}{lowers[0]}")
	elif "?" in lowers:
		lowers.remove("?")
		invalid_uppers.remove(upper)
		for lower in lowers:
			invalid_lowers[upper].remove(lower)
		element_re.append(f"(({upper}[{''.join(lowers)}]?)(?![{''.join(invalid_lowers[upper])}]))(?![a-z])")
		invalid_re.append(f"{upper}[{''.join(lowers)}][{''.join(invalid_lowers[upper])}]+")
		invalid_re.append(f"{upper}[{''.join(invalid_lowers[upper])}]+")
	else:
		invalid_uppers.remove(upper)
		for lower in lowers:
			invalid_lowers[upper].remove(lower)
		element_re.append(f"({upper}[{''.join(lowers)}]?)(?![a-z])")
		invalid_re.append(f"{upper}[{''.join(lowers)}][{''.join(invalid_lowers[upper])}]+")
		invalid_re.append(f"{upper}[{''.join(invalid_lowers[upper])}]+")

# print(invalid_uppers, invalid_lowers)

element_re.sort()

# Isotopes either as [C12], C[12] or [12C]
isotopes_re = []
for elem in element_re:
	isotopes_re.append(rf"\b{elem}\[[0-9]+\]")
	isotopes_re.append(rf"\[{elem}[0-9]+\]")
	isotopes_re.append(rf"\[[0-9]+{elem}\]")


@lru_cache()
def _get_formula_parser():
	"""
	Create a forward pyparsing parser for chemical formulae

	BNF for simple chemical formula (no nesting)

		integer :: '0'..'9'+
		element :: 'A'..'Z' 'a'..'z'*
		term :: element [integer]
		formula :: term+


	BNF for nested chemical formula

		integer :: '0'..'9'+
		element :: 'A'..'Z' 'a'..'z'*
		term :: (element | '(' formula ')') [integer]
		formula :: term+

	Notes
	-----
	Based on http://stackoverflow.com/a/18555142/790973
	Copyright 2013 Paul McGuire (http://stackoverflow.com/users/165216/paul-mcguire)
	Licensed under CC-BY-SA 3.0.
	"""

	Forward, Group, OneOrMore = pyparsing.Forward, pyparsing.Group, pyparsing.OneOrMore
	Optional, ParseResults = pyparsing.Optional, pyparsing.ParseResults
	Suppress, Word, nums = pyparsing.Suppress, pyparsing.Word, pyparsing.nums

	LPAR, RPAR = map(Suppress, "()")
	integer = Word(nums)

	# add parse action to convert integers to ints, to support doing addition
	# and multiplication at parse time
	integer.setParseAction(lambda t: int(t[0]))

	element = pyparsing.Regex('|'.join(isotopes_re + element_re) + "$")

	# element = pyparsing.Regex(
	# 		r"A[cglmrstu]|B[aehikr]?|C[adeflmorsu]?|D[bsy]|E[rsu]|F[emr]?|G[ade]"
	# 		r"|H[efgos]?|I[nr]?|Kr?|L[airu]|M[dgnot]|N[abdeiop]?|Os?|P[abdmortu]?"
	# 		r"|R[abefghnu]|S[bcegimnr]?|T[abcehilm]|Uu[bhopqst]|U|V|W|Xe|Yb?|Z[nr]")

	# forward declare 'formula' so it can be used in definition of 'term'
	formula = Forward()

	term = Group(
			(element | Group(LPAR + formula + RPAR)("subgroup"))
			+ Optional(integer, default=1)("mult")
			)

	# add parse actions for parse-time processing

	# parse action to multiply out subgroups
	def multiplyContents(tokens):
		t = tokens[0]
		# if these tokens contain a subgroup, then use multiplier to
		# extend counts of all elements in the subgroup
		if t.subgroup:
			mult = t.mult
			for term in t.subgroup:
				term[1] *= mult
			return t.subgroup

	term.setParseAction(multiplyContents)

	# add parse action to sum up multiple references to the same element
	def sum_by_element(tokens):
		elementsList = [t[0] for t in tokens]

		# construct set to see if there are duplicates
		duplicates = len(elementsList) > len(set(elementsList))

		# if there are duplicate element names, sum up by element and
		# return a new nested ParseResults
		if duplicates:
			ctr = defaultdict(int)
			for t in tokens:
				ctr[t[0]] += t[1]
			return ParseResults([ParseResults([k, v]) for k, v in ctr.items()])

	# define contents of a formula as one or more terms
	formula << OneOrMore(term)
	formula.setParseAction(sum_by_element)

	return formula


def _parse_stoich(stoich):
	if stoich == 'e':  # special case, the electron is not an element
		return {}

	symbols = ELEMENTS.symbols + ["D", "T"]

	if re.findall('|'.join(invalid_re), stoich):
		raise ValueError(f"Unrecognised formula: {stoich}")

	return {
			symbols.index(k) + 1: n for k, n
			in _get_formula_parser().parseString(stoich)
			}


def string_to_composition(
		formula, prefixes=None,
		suffixes=('(s)', '(l)', '(g)', '(aq)')):
	"""
	Parse composition of formula representing a chemical formula

	Composition is represented as a dict mapping int -> int (atomic
	number -> multiplicity). "Atomic number" 0 represents net charge.

	Parameters
	----------
	formula: str
		Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
	prefixes: iterable strings
		Prefixes to ignore, e.g. ('.', 'alpha-')
	suffixes: tuple of strings
		Suffixes to ignore, e.g. ('(g)', '(s)')

	Examples
	--------
	>>> string_to_composition('NH4+') == {0: 1, "H": 4, "N": 1}
	True
	>>> string_to_composition('.NHO-(aq)') == {0: -1, "H": 1, "N": 1, "O": 1}
	True
	>>> string_to_composition('Na2CO3.7H2O') == {"Na": 2, "C": 1, "O": 10, "H": 14}
	True

	"""
	if prefixes is None:
		prefixes = _latex_mapping.keys()
	stoich_tok, chg_tok = _formula_to_parts(formula, prefixes, suffixes)[:2]
	tot_comp = {}
	parts = stoich_tok.split('.')
	for idx, stoich in enumerate(parts):
		if idx == 0:
			m = 1
		else:
			m, stoich = _get_leading_integer(stoich)

		# comp = _parse_stoich(stoich)
		if stoich == 'e':  # special case, the electron is not an element
			comp = {}
		else:
			try:
				if re.findall('|'.join(invalid_re), stoich):
					raise ValueError(f"Unrecognised formula: {formula}")

				comp = _get_formula_parser().parseString(stoich)
			except pyparsing.ParseException:
				raise ValueError(f"Unrecognised formula: {formula}")

		# for k, v in comp.items():
		for k, v in comp:
			if k not in tot_comp:
				tot_comp[k] = m * v
			else:
				tot_comp[k] += m * v
	if chg_tok is not None:
		tot_comp[0] = _get_charge(chg_tok)

	return tot_comp


def to_reaction(line, substance_keys, token, cls, globals_=None, **kwargs):
	"""
	Parses a string into a Reaction object and substances

	Reac1 + 2 Reac2 + (2 Reac1) -> Prod1 + Prod2; 10**3.7; ref='doi:12/ab'
	Reac1 = Prod1; 2.1;

	Parameters
	----------
	line: str
		string representation to be parsed
	substance_keys: iterable of strings
		Allowed names, e.g. ('H2O', 'H+', 'OH-')
	token : str
		delimiter token between reactant and product side
	cls : class
		e.g. subclass of Reaction
	globals_: dict (optional)
		Globals passed on to :func:`eval`, when ``None``:
		`chempy.units` is used with 'chempy'
		and 'default_units' extra entries.

	Notes
	-----
	This function calls :func:`eval`, hence there are severe security concerns
	with running this on untrusted data.

	"""

	parts = line.rstrip('\n').split(';')
	stoich = parts[0].strip()
	if len(parts) > 2:
		kwargs.update(eval('dict(' + ';'.join(parts[2:]) + '\n)', globals_ or {}))
	if len(parts) > 1:
		param = parts[1].strip()
	else:
		param = kwargs.pop('param', 'None')

	if isinstance(param, str):
		param = None if globals_ is False else eval(param, globals_)

	if token not in stoich:
		raise ValueError("Missing token: %s" % token)

	reac_prod = [[y.strip() for y in x.split(' + ')] for x in stoich.split(token)]

	act, inact = [], []
	for elements in reac_prod:
		act.append(_parse_multiplicity([x for x in elements if not x.startswith('(')], substance_keys))
		inact.append(_parse_multiplicity(
				[x[1:-1] for x in elements if x.startswith('(') and x.endswith(')')],
				substance_keys
				))

	# stoich coeff -> dict
	return cls(act[0], act[1], param, inact_reac=inact[0], inact_prod=inact[1], **kwargs)


def mass_from_composition(composition, charge=0):
	"""
	Calculates molecular mass from atomic weights

	Parameters
	----------
	composition: dict
		Dictionary mapping str or int (element symbol or atomic number) to int (coefficient)
	charge:
		The charge of the composition. Can also be given as the "0" key of ``composition``

	Returns
	-------
	float
		molecular weight in atomic mass units


	Notes
	-----
	Atomic number 0 denotes charge or "net electron defficiency"

	Examples
	--------
	>>> f'{mass_from_composition({0: -1, "H": 1, 8: 1}):.2f}'
	'17.01'
	"""

	if charge and 0 in composition:
		if charge != composition[0]:
			raise ValueError(
					"'charge' can only be specified once, "
					"either as a keyword argument or as the '0' key of 'composition'")

	mass = 0.0
	for k, v in composition.items():
		if k == 0:  # electron
			mass -= v * 5.489e-4
		elif isinstance(k, str):
			mass += v * ELEMENTS[k].mass
		elif isinstance(k, int):
			mass += v * ELEMENTS[k].mass
	return mass



#!/usr/bin/env python
#
#  compound.py
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# 3rd party
from domdf_python_tools.bases import Dictable
from memoized_property import memoized_property

# this package
from chemistry_tools.pubchem.atom import parse_atoms
from chemistry_tools.pubchem.bond import parse_bonds
from chemistry_tools.pubchem.enums import CoordinateType
from chemistry_tools.pubchem.full_record import parse_full_record, rest_get_full_record
from chemistry_tools.pubchem.properties import (
	_force_valid_properties, insert_valid_properties_table, parse_properties,
	rest_get_properties_json, valid_properties,
	)
from chemistry_tools.pubchem.synonyms import get_synonyms


class Compound(Dictable):
	"""
	Corresponds to a single record from the PubChem Compound database.

	The PubChem Compound database is constructed from the Substance database
	using a standardization and deduplication process.
	Each Compound is uniquely identified by a CID.
	"""

	def __init__(self, title, CID, description, **_):
		"""
		Initialize with a record dict from the PubChem PUG REST service.
		"""

		super().__init__()

		self.title = str(title)
		self.CID = int(CID)
		self.description = str(description)
		self._properties = {prop: None for prop in valid_properties}
		self.record_type = "2d"
		self._synonyms = None

		# Pre-cache all properties
		# self.get_properties("all")

		self._has_full_record = False

	@property
	def __dict__(self):
		return dict(
				title=self.title,
				CID=self.CID,
				description=self.description,
				properties=self._properties,
				record_type=self.record_type,
				counts=self._record["counts"],
				atoms=self._atoms,
				bonds=self._bonds,
				)

	def __repr__(self):
		return f'Compound({self.cid})' if self.cid else 'Compound()'

	def to_series(self):
		"""
		Return a pandas :class:`~pandas.Series` containing Compound data.
		"""

		import pandas as pd
		return pd.Series(dict(self))

	@property
	def cid(self):
		return self.CID

	@property
	def has_full_record(self):
		return self._has_full_record

	@memoized_property
	def _record(self):

		# Only requested when required
		record = parse_full_record(rest_get_full_record(self.CID, "cid", self.record_type))[0]
		self._has_full_record = True

		for prop in record["properties"]:
			if not self._properties["CanonicalSMILES"]:
				if prop.label == "SMILES" and prop.name == "Canonical":
					self._properties["CanonicalSMILES"] = prop.value

			if not self._properties["IsomericSMILES"]:
				if prop.label == "SMILES" and prop.name == "Isomeric":
					self._properties["IsomericSMILES"] = prop.value

			if not self._properties["TPSA"]:
				if prop.label == "Topological" and prop.name == "Polar Surface Area":
					self._properties["TPSA"] = prop.value

			if not self._properties["InChIKey"]:
				if prop.label == "InChIKey" and prop.name == "Standard":
					self._properties["InChIKey"] = prop.value

			if not self._properties["InChI"]:
				if prop.label == "InChI" and prop.name == "Standard":
					self._properties["InChI"] = prop.value

			if not self._properties["HBondAcceptorCount"]:
				if prop.label == "Count" and prop.name == "Hydrogen Bond Acceptor":
					self._properties["HBondAcceptorCount"] = prop.value

			if not self._properties["HBondDonorCount"]:
				if prop.label == "Count" and prop.name == "Hydrogen Bond Donor":
					self._properties["HBondDonorCount"] = prop.value

			if not self._properties["RotatableBondCount"]:
				if prop.label == "Count" and prop.name == "Rotatable Bond":
					self._properties["RotatableBondCount"] = prop.value

			if not self._properties["MolecularWeight"]:
				if prop.label == "Molecular Weight" and prop.name is None:
					self._properties["MolecularWeight"] = prop.value

			# TODO: label='Weight', name='MonoIsotopic',
				#label='Molecular Formula', name=None,
				#label='Mass', name='Exact',
				# label='Log P', name='XLogP3'
				#
		return record

	@memoized_property
	def _atoms(self):
		"""
		Derive Atom objects from the record.
		"""

		if 'atoms' not in self._record:
			return

		atoms_dict = self._record['atoms']
		coords_dict = self._record.get('coords', None)
		return parse_atoms(atoms_dict, coords_dict)

	@memoized_property
	def _bonds(self):
		"""
		Derive Bond objects from the record.
		"""


		if 'bonds' not in self._record:
			return

		bonds_dict = self._record['bonds']
		coords_dict = self._record.get('coords', None)
		return parse_bonds(bonds_dict, coords_dict)

	def precache(self):
		self.get_properties("all")
		_ = self._atoms
		_ = self._bonds

	@property
	def atoms(self):
		"""
		List of :class:`Atoms <chemistry_tools.pubchem.atom.Atom>`
		in this Compound.
		"""

		return sorted(self._atoms.values(), key=lambda x: x.aid)

	@property
	def bonds(self):
		"""
		List of :class:`Bonds <chemistry_tools.pubchem.bond.Bond>`
		between :class:`Atoms <chemistry_tools.pubchem.atom.Atom>`
		in this Compound.
		"""

		return sorted(self._bonds.values(), key=lambda x: (x.aid1, x.aid2))

	@property
	def coordinate_type(self):
		if CoordinateType.TWO_D in self._record['coords'][0]['type']:
			return '2d'
		elif CoordinateType.THREE_D in self._record['coords'][0]['type']:
			return '3d'

	@property
	def elements(self):
		"""List of element symbols for atoms in this Compound."""
		return [a.element for a in self.atoms]

	@insert_valid_properties_table()
	def get_properties(self, properties):
		"""
		Returns the requested properties for the Compound

		:param properties: The properties to retrieve for the compound. See the table below. Can be either a comma-separated string or a list.

		:: See chemistry_tools.pubchem.properties.valid_property_descriptions for a list of valid properties ::

		:type properties: str, Sequence[str]

		:return:
		:rtype:
		"""

		if isinstance(properties, str) and properties.lower() == "all":
			properties = list(valid_properties.keys())

		properties = _force_valid_properties(properties)

		cached_properties = []
		properties_to_get = []

		for prop in properties:
			if self._properties[prop] is not None:
				cached_properties.append(prop)
			else:
				properties_to_get.append(prop)

		output = {}

		if properties_to_get:
			print("Getting from API")
			data = rest_get_properties_json(self.CID, "cid", properties)
			new_properties = parse_properties(data)[0]

			for prop in properties_to_get:
				self._properties[prop] = new_properties[prop]
				output[prop] = new_properties[prop]

		for prop in cached_properties:
			print("Getting from cache")
			output[prop] = self._properties[prop]

		return output

	@insert_valid_properties_table()
	def get_property(self, prop):
		"""
		Get a single property for the compound

		:param prop: The property to retrieve for the compound. See the table below.

		:: See chemistry_tools.pubchem.properties.valid_property_descriptions for a list of valid properties ::

		:type prop: str

		:return:
		:rtype:
		"""

		prop = str(prop)

		if prop not in self._properties:
			raise ValueError(f"Unknown property '{prop}'")

		if self._properties[prop] is not None:
			print("Getting from cache")
			return self._properties[prop]
		else:
			print("Getting from API")
			data = rest_get_properties_json(self.CID, "cid", prop)
			new_properties = parse_properties(data)[0]

			self._properties[prop] = new_properties[prop]
			return new_properties[prop]

	@property
	def synonyms(self):
		"""
		Returns a list of synonyms for the Compound.

		:rtype: List[str]
		"""

		if not self._synonyms:
			data = get_synonyms(self.CID, "cid")[0]

			if not data["CID"] == self.CID:
				raise ValueError("Wrong compound returned")

			else:
				self._synonyms = data["synonyms"]

		return self._synonyms

	@classmethod
	def from_cid(cls, cid, record_type="2d"):
		"""
		Returns the Compound objects for the compound with the given CID

		:return:
		:rtype:
		"""

		from chemistry_tools.pubchem.lookup import get_compounds

		comp = get_compounds(cid, "cid")[0]

		if comp.CID != int(cid):
			raise ValueError("Wrong compound returned")

		comp.record_type = record_type

		return comp

	# Convenience attributes for some properties

	@property
	def molecular_formula(self):
		"""
		Molecular formula.
		"""

		return self.get_property("MolecularFormula")

	@property
	def canonical_smiles(self):
		"""
		Canonical SMILES, with no stereochemistry information.
		"""

		return self.get_property("CanonicalSMILES")

	smiles = canonical_smiles

	@property
	def charge(self):
		return self.get_property("Charge")


	@property
	def molecular_weight(self):
		"""
		Molecular Weight.
		"""

		return self.get_property("MolecularWeight")

	molecular_mass = molecular_weight

	@memoized_property
	def canonicalized(self):
		for prop in self._record["properties"]:
			if prop.label == "Compound" and prop.name == "Canonicalized":
				return bool(prop.value)

		return False

	def get_iupac_name(self, type_="Systematic"):
		# Allowed, CAS-like Style, Markup, Preferred, Systematic, Traditional
		for prop in self._record["properties"]:
			if prop.label == "IUPAC Name" and prop.name == type_.capitalize():
				return prop.value

		return False

	@memoized_property
	def iupac_name(self):
		return self.get_iupac_name("Preferred")

	@memoized_property
	def systematic_name(self):
		return self.get_iupac_name("Systematic")

	@property
	def fingerprint(self):
		"""
		Raw padded and hex-encoded fingerprint, as returned by the PUG REST API.
		"""

		for prop in self._record["properties"]:
			if prop.label == "Fingerprint" and prop.name == "SubStructure Keys":
				return prop.value

	@property
	def cactvs_fingerprint(self):
		"""
		PubChem CACTVS fingerprint.

		Each bit in the fingerprint represents the presence or absence of one of 881 chemical substructures.

		More information at ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
		"""

		# Skip first 4 bytes (contain length of fingerprint) and last 7 bits (padding) then re-pad to 881 bits

		return '{:020b}'.format(int(self.fingerprint[8:], 16))[:-7].zfill(881)

	# @memoized_property
	# def hill_formula(self):
	# 	element_count = Counter(self.elements)
	# 	hill = []
	#
	# 	alphabet = sorted(ELEMENTS.symbols)
	#
	# 	if "C" in element_count:
	# 		hill.append("C")
	# 		alphabet.remove("C")
	# 		count = element_count["C"]
	# 		if count > 1:
	# 			hill.append(f"<sub>{count}</sub>")
	# 		if "H" in element_count:
	# 			hill.append("H")
	# 			alphabet.remove("H")
	# 			count = element_count["H"]
	# 			if count > 1:
	# 				hill.append(f"<sub>{count}</sub>")
	#
	# 	for element in alphabet:
	# 		if element in element_count:
	# 			hill.append(element)
	# 			count = element_count[element]
	# 			if count > 1:
	# 				hill.append(f"<sub>{count}</sub>")
	#
	# 	return ''.join(hill)

# TODO from record:
	# charge
	# properties
		# label='Compound', name='Canonicalized'
		# label='Compound Complexity', name=None
	# cid
	# counts


# TODO:
def compounds_to_frame(compounds):
	"""
	Construct a pandas :class:`~pandas.DataFrame` from a list of :class:`~pubchempy.Compound` objects.
	"""

	import pandas as pd
	if isinstance(compounds, Compound):
		compounds = [compounds]
	return pd.DataFrame.from_records([dict(c) for c in compounds], index='CID')



#!/usr/bin/env python3
#
#  utils.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |


# stdlib
import re
from functools import lru_cache

# this package
from chemistry_tools.elements import ELEMENTS

#  3rd party


# Common chemical groups
GROUPS = {
		'Abu': 'C4H7NO',
		'Acet': 'C2H3O',
		'Acm': 'C3H6NO',
		'Adao': 'C10H15O',
		'Aib': 'C4H7NO',
		'Ala': 'C3H5NO',
		'Arg': 'C6H12N4O',
		'Argp': 'C6H11N4O',
		'Asn': 'C4H6N2O2',
		'Asnp': 'C4H5N2O2',
		'Asp': 'C4H5NO3',
		'Aspp': 'C4H4NO3',
		'Asu': 'C8H13NO3',
		'Asup': 'C8H12NO3',
		'Boc': 'C5H9O2',
		'Bom': 'C8H9O',
		'Bpy': 'C10H8N2',  # Bipyridine
		'Brz': 'C8H6BrO2',
		'Bu': 'C4H9',
		'Bum': 'C5H11O',
		'Bz': 'C7H5O',
		'Bzl': 'C7H7',
		'Bzlo': 'C7H7O',
		'Cha': 'C9H15NO',
		'Chxo': 'C6H11O',
		'Cit': 'C6H11N3O2',
		'Citp': 'C6H10N3O2',
		'Clz': 'C8H6ClO2',
		'Cp': 'C5H5',
		'Cy': 'C6H11',
		'Cys': 'C3H5NOS',
		'Cysp': 'C3H4NOS',
		'Dde': 'C10H13O2',
		'Dnp': 'C6H3N2O4',
		'Et': 'C2H5',
		'Fmoc': 'C15H11O2',
		'For': 'CHO',
		'Gln': 'C5H8N2O2',
		'Glnp': 'C5H7N2O2',
		'Glp': 'C5H5NO2',
		'Glu': 'C5H7NO3',
		'Glup': 'C5H6NO3',
		'Gly': 'C2H3NO',
		'Hci': 'C7H13N3O2',
		'Hcip': 'C7H12N3O2',
		'His': 'C6H7N3O',
		'Hisp': 'C6H6N3O',
		'Hser': 'C4H7NO2',
		'Hserp': 'C4H6NO2',
		'Hx': 'C6H11',
		'Hyp': 'C5H7NO2',
		'Hypp': 'C5H6NO2',
		'Ile': 'C6H11NO',
		'Ivdde': 'C14H21O2',
		'Leu': 'C6H11NO',
		'Lys': 'C6H12N2O',
		'Lysp': 'C6H11N2O',
		'Mbh': 'C15H15O2',
		'Me': 'CH3',
		'Mebzl': 'C8H9',
		'Meobzl': 'C8H9O',
		'Met': 'C5H9NOS',
		'Mmt': 'C20H17O',
		'Mtc': 'C14H19O3S',
		'Mtr': 'C10H13O3S',
		'Mts': 'C9H11O2S',
		'Mtt': 'C20H17',
		'Nle': 'C6H11NO',
		'Npys': 'C5H3N2O2S',
		'Nva': 'C5H9NO',
		'Odmab': 'C20H26NO3',
		'Orn': 'C5H10N2O',
		'Ornp': 'C5H9N2O',
		'Pbf': 'C13H17O3S',
		'Pen': 'C5H9NOS',
		'Penp': 'C5H8NOS',
		'Ph': 'C6H5',
		'Phe': 'C9H9NO',
		'Phepcl': 'C9H8ClNO',
		'Phg': 'C8H7NO',
		'Pmc': 'C14H19O3S',
		'Ppa': 'C8H7O2',
		'Pro': 'C5H7NO',
		'Prop': 'C3H7',
		'Py': 'C5H5N',
		'Pyr': 'C5H5NO2',
		'Sar': 'C3H5NO',
		'Ser': 'C3H5NO2',
		'Serp': 'C3H4NO2',
		'Sta': 'C8H15NO2',
		'Stap': 'C8H14NO2',
		'Tacm': 'C6H12NO',
		'Tbdms': 'C6H15Si',
		'Tbu': 'C4H9',
		'Tbuo': 'C4H9O',
		'Tbuthio': 'C4H9S',
		'Tfa': 'C2F3O',
		'Thi': 'C7H7NOS',
		'Thr': 'C4H7NO2',
		'Thrp': 'C4H6NO2',
		'Tips': 'C9H21Si',
		'Tms': 'C3H9Si',
		'Tos': 'C7H7O2S',
		'Trp': 'C11H10N2O',
		'Trpp': 'C11H9N2O',
		'Trt': 'C19H15',
		'Tyr': 'C9H9NO2',
		'Tyrp': 'C9H8NO2',
		'Val': 'C5H9NO',
		'Valoh': 'C5H9NO2',
		'Valohp': 'C5H8NO2',
		'Xan': 'C13H9O',
		}


_isotope_regex_1 = re.compile(r"^([A-z]+)(\[\d*\])$")
_isotope_regex_2 = re.compile(r"^(\[[A-z]+)(\d*\])$")
_isotope_regex_3 = re.compile(r"^(\[\d*)([A-z]+\])$")
_iso_bracket_regex = re.compile(r"^(\[)(\d+)(\])$")


@lru_cache()
def _split_isotope(string):
	"""
	Returns the symbol and massnumber for the isotope represented by ``string``.

	Valid isotopes include [C12], C[12] and [12C]

	:param string:
	:type string: str

	:return:
	:rtype: (str, int)
	"""

	isotope = 0

	iso_re_1 = _isotope_regex_1.findall(string)
	iso_re_2 = _isotope_regex_2.findall(string)
	iso_re_3 = _isotope_regex_3.findall(string)

	if iso_re_1:
		elem, isotope = iso_re_1[0]
		isotope = _iso_bracket_regex.findall(isotope)[0][1]
	elif iso_re_2:
		elem, isotope = iso_re_2[0]
		elem = elem.lstrip("[")
		isotope = isotope.rstrip("]")
	elif iso_re_3:
		isotope, elem = iso_re_3[0]
		isotope = isotope.lstrip("[")
		elem = elem.rstrip("]")
	else:
		elem = string

	if elem not in ELEMENTS:
		raise ValueError(f'Unknown chemical element with symbol {elem}')

	return ELEMENTS[elem].symbol, int(isotope)


def _hill_order(*symbols):
	"""
	Return iterator over element symbols in order of Hill notation.

	Examples
	--------
	>>> for i in _hill_order('H', 'C[12]', 'O'): print(i, end='')
	CHO

	"""

	if len(symbols) == 0:
		raise TypeError("'_hill_order' requires at least 1 argument (0 given)")
	elif len(symbols) == 1:
		if isinstance(symbols[0], (list, tuple, set)):
			symbols = symbols[0]

	symbols = list(set(symbols))

	isotope_re = r"^(%s)(\[[0-9]*\])?$"
	carbon_re = re.compile(isotope_re % "C")
	hydrogen_re = re.compile(isotope_re % "H")

	carbon_isotopes = list(filter(carbon_re.findall, symbols))

	if carbon_isotopes:
		for isotope in sorted(carbon_isotopes):
			symbols.remove(isotope)
			yield isotope

		hydrogen_isotopes = list(filter(hydrogen_re.findall, symbols))
		for isotope in sorted(hydrogen_isotopes):
			symbols.remove(isotope)
			yield isotope

	yield from sorted(symbols)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  _table.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#

__all__ = ("PERIODS", "BLOCKS", "GROUPS", "SERIES")


PERIODS = {1: 'K', 2: 'L', 3: 'M', 4: 'N', 5: 'O', 6: 'P', 7: 'Q'}

BLOCKS = {'s': '', 'g': '', 'f': '', 'd': '', 'p': ''}

GROUPS = {
		1: ('IA', 'Alkali metals'),
		2: ('IIA', 'Alkaline earths'),
		3: ('IIIB', ''),
		4: ('IVB', ''),
		5: ('VB', ''),
		6: ('VIB', ''),
		7: ('VIIB', ''),
		8: ('VIIIB', ''),
		9: ('VIIIB', ''),
		10: ('VIIIB', ''),
		11: ('IB', 'Coinage metals'),
		12: ('IIB', ''),
		13: ('IIIA', 'Boron group'),
		14: ('IVA', 'Carbon group'),
		15: ('VA', 'Pnictogens'),
		16: ('VIA', 'Chalcogens'),
		17: ('VIIA', 'Halogens'),
		18: ('VIIIA', 'Noble gases'),
		}

SERIES = {
		1: 'Nonmetals',
		2: 'Noble gases',
		3: 'Alkali metals',
		4: 'Alkaline earth metals',
		5: 'Metalloids',
		6: 'Halogens',
		7: 'Poor metals',
		8: 'Transition metals',
		9: 'Lanthanides',
		10: 'Actinides',
		}



# This file is managed by `git_helper`. Don't edit it directly
# Copyright (C) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This script based on https://github.com/rocky/python-uncompyle6/blob/master/__pkginfo__.py

import pathlib

__all__ = [
		"__copyright__",
		"__version__",
		"modname",
		"pypi_name",
		"py_modules",
		"entry_points",
		"__license__",
		"short_desc",
		"author",
		"author_email",
		"github_username",
		"web",
		"github_url",
		"project_urls",
		"repo_root",
		"long_description",
		"install_requires",
		"extras_require",
		"classifiers",
		"keywords",
		"import_name",
		]

__copyright__ = """
2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
"""

__version__ = "0.3.0"

modname = "chemistry_tools"
pypi_name = "chemistry_tools"
import_name = "chemistry_tools"
py_modules = []
entry_points = {
		"console_scripts": []
		}

__license__ = "GNU Lesser General Public License v3 or later (LGPLv3+)"

short_desc = "Python tools for analysis of chemical compounds"

__author__ = author = "Dominic Davis-Foster"
author_email = "dominic@davis-foster.co.uk"
github_username = "domdfcoding"
web = github_url = f"https://github.com/domdfcoding/chemistry_tools"
project_urls = {
		"Documentation": f"https://chemistry_tools.readthedocs.io",  # TODO: Make this link match the package version
		"Issue Tracker": f"{github_url}/issues",
		"Source Code": github_url,
		}

repo_root = pathlib.Path(__file__).parent

# Get info from files; set: long_description
long_description = (repo_root / "README.rst").read_text().replace("0.3.0", __version__) + '\n'
conda_description = """Python tools for analysis of chemical compounds


Before installing please ensure you have added the following channels: domdfcoding, conda-forge"""
__all__.append("conda_description")

install_requires = (repo_root / "requirements.txt").read_text().split('\n')
extras_require = {'pubchem': ['tabulate>=0.8.3', 'pandas>=1.0.1', 'memoized-property>=1.0.3', 'domdf_python_tools>=0.3.3', 'Pillow>=7.0.0'], 'elements': ['domdf_python_tools>=0.3.3', 'memoized-property>=1.0.3'], 'formulae': ['quantities>=0.12.4', 'domdf_python_tools>=0.3.3', 'tabulate>=0.8.3', 'pyparsing>=2.2.0', 'pandas>=1.0.1', 'mathematical>=0.1.7', 'cawdrey>=0.1.2', 'memoized-property>=1.0.3'], 'plotting': ['matplotlib>=3.0.0'], 'toxnet': ['beautifulsoup4>=4.7.0'], 'all': ['Pillow>=7.0.0', 'beautifulsoup4>=4.7.0', 'cawdrey>=0.1.2', 'domdf_python_tools>=0.3.3', 'mathematical>=0.1.7', 'matplotlib>=3.0.0', 'memoized-property>=1.0.3', 'pandas>=1.0.1', 'pyparsing>=2.2.0', 'quantities>=0.12.4', 'tabulate>=0.8.3']}

classifiers = [
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Intended Audience :: Education',
		'Intended Audience :: Science/Research',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Topic :: Database :: Front-Ends',
		'Topic :: Scientific/Engineering :: Bio-Informatics',
		'Topic :: Scientific/Engineering :: Chemistry',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: Implementation :: CPython',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3 :: Only',
		'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',

		]

keywords = ""



#!/usr/bin/env python3
#
#  enums.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from domdf_python_tools.enums import StrEnum


class PubChemNamespace(StrEnum):
	cid = "cid"
	name = "name"
	smiles = "smiles"
	# inchi = "inchi"# TODO: requires argument
	# sdf = "sdf"
	inchikey = "inchikey"
	# formula = "formula"
	Cid = "cid"
	Name = "name"
	Smiles = "smiles"
	# Inchi = "inchi"# TODO: requires argument
	# Sdf = "sdf"
	Inchikey = "inchikey"
	# Formula = "formula"
	CID = "cid"
	NAME = "name"
	SMILES = "smiles"
	# INCHI = "inchi"# TODO: requires argument
	# SDF = "sdf"
	INCHIKEY = "inchikey"
	# FORMULA = "formula"

	# TODO: listkey for formula lookup https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest$_Toc494865583

	@classmethod
	def is_valid_value(cls, value):
		return str(value) in set(str(item) for item in PubChemNamespace)


class PubChemFormats(StrEnum):
	JSON = "JSON"
	Json = "JSON"
	json = "JSON"
	XML = "XML"
	Xml = "XML"
	xml = "XML"
	CSV = "CSV"
	csv = "CSV"
	Csv = "CSV"
	PNG = "PNG"
	png = "PNG"
	Png = "PNG"

	@classmethod
	def is_valid_value(cls, value):
		return str(value).upper() in set(str(item) for item in PubChemFormats)


class CoordinateType:
	TWO_D = 1
	THREE_D = 2
	SUBMITTED = 3
	EXPERIMENTAL = 4
	COMPUTED = 5
	STANDARDIZED = 6
	AUGMENTED = 7
	ALIGNED = 8
	COMPACT = 9
	UNITS_ANGSTROMS = 10
	UNITS_NANOMETERS = 11
	UNITS_PIXEL = 12
	UNITS_POINTS = 13
	UNITS_STDBONDS = 14
	UNITS_UNKNOWN = 255


# -*- coding: utf-8 -*-
"""
test_download
~~~~~~~~~~~~~

Test downloading.

"""

import os
import csv
import shutil
import tempfile

import pytest

from chemistry_tools.pubchem.images import get_structure_image
from chemistry_tools.pubchem.properties import rest_get_properties


@pytest.fixture(scope='module')
def tmp_dir():
	tmpdir = tempfile.mkdtemp()
	yield tmpdir
	shutil.rmtree(tmpdir)


def test_image_download(tmp_dir):
	img = get_structure_image("Asprin")
	img.save(os.path.join(tmp_dir, 'aspirin.png'))


def test_csv_download(tmp_dir):
	csv_content = rest_get_properties([1, 2, 3], namespace="cid", properties="CanonicalSMILES,IsomericSMILES", format_="csv")
	with open(os.path.join(tmp_dir, 's.csv'), "w") as fp:
		fp.write(csv_content)

	with open(os.path.join(tmp_dir, 's.csv')) as f:
		rows = list(csv.reader(f))
		assert rows[0] == ['CID', 'CanonicalSMILES', 'IsomericSMILES']
		assert rows[1][0] == '1'
		assert rows[2][0] == '2'
		assert rows[3][0] == '3'
